/*
  Copyright (C) 2006 QLogic Corporation, All rights reserved.
  Copyright (c) 2004,2005. PathScale, Inc. All rights reserved.
    
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions
  are met:
    
  + All copies of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer. 

  + Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

  + If user makes modifications to this software, and provides those 
    modifications to PathScale, user assigns to PathScale, Inc. ownership
    of the provided modifications and all intellectual property rights 
    embodied in those modifications. 
    
  + The name of PathScale, Inc. may not be used to endorse or promote 
    products derived from this software without specific prior written 
    permission. 
      
  THIS SOFTWARE IS PROVIDED BY PATHSCALE, INC. AND ITS LICENSORS "AS IS" 
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL PATHSCALE, INC. OR ITS 
  LICENSORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/
/*
 * Derived from mpi_bandwidth.c and mpi_latency.c
 * Copyright (C) 2002-2003 the Network-Based Computing Laboratory
 * (NBCL), The Ohio State University.  
 *  http://nowlab.cis.ohio-state.edu/projects/mpi-iba/
 */


/* Unlike mpi_bandwidth, this test uses blocking sends and receives. */

#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#define MYBUFSIZE 4*1024*1025
char s_buf[MYBUFSIZE];
char r_buf[MYBUFSIZE];
int skip = 10;
double *timings = 0;

#define NUMBUCKETS 10

int main(int argc,char *argv[])
{
    int myid, numprocs, i, l;
    int size, loop, loopcnt=1;
    MPI_Status stat;
    double t_start=0.0, t_end=0.0;

    MPI_Init(&argc,&argv);
    MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
    MPI_Comm_rank(MPI_COMM_WORLD,&myid);

    while((i=getopt(argc,argv,"+r:")) > 0)
    switch(i) {
    case 'r':
        loopcnt = strtoul(optarg, NULL, 0);
        break;
    default:
        goto usage;
        break;
    }

    if(loopcnt<1 || optind != (argc-2)) {
usage:
       fprintf(stderr, "Usage: mpi_bw [-r repeatcnt] loop msg_size\n");
       MPI_Finalize();
       return 0;
    }    
    loop = atoi(argv[optind]);
    size=atoi(argv[optind+1]);
    if(size > MYBUFSIZE){
            fprintf(stderr, "Maximum message size is %d\n",MYBUFSIZE);
            MPI_Finalize();
            return 0;
    }

    if(numprocs != 2) { /* after other checks to be most helpful */
        printf("Expected exactly 2 MPI processes\n");
        MPI_Abort( MPI_COMM_WORLD, 1 );
    }

    /* touch the data */
    for ( i=0; i<size; i++ ){
        if(myid == 0) {
        s_buf[i]='a';
        r_buf[i]='b';
        }
        else {
        s_buf[i]='B';
        r_buf[i]='A';
        }
    }

    for(l=0; l<loopcnt; l++) {
        MPI_Barrier( MPI_COMM_WORLD);
        if (myid == 0)
        {
            for ( i=0; i< skip; i++ ) {
                MPI_Send(s_buf, size, MPI_CHAR, 1, i, MPI_COMM_WORLD);
            }
            MPI_Recv(NULL, 0, MPI_CHAR, 1, i + 1000, MPI_COMM_WORLD, &stat);
            t_start=MPI_Wtime();
            for ( i=0; i< loop; i++ ) {
                MPI_Send(s_buf, size, MPI_CHAR, 1, i, MPI_COMM_WORLD);
            }
            MPI_Recv(NULL, 0, MPI_CHAR, 1, i + 1000, MPI_COMM_WORLD, &stat);
            t_end=MPI_Wtime();
        }else{
            for ( i=0; i< skip; i++ ) {
                MPI_Recv(r_buf, size, MPI_CHAR, 0, i, MPI_COMM_WORLD, &stat);
            }
            MPI_Send(NULL, 0, MPI_CHAR, 0, i + 1000, MPI_COMM_WORLD);
            for ( i=0; i< loop; i++ ) {
                MPI_Recv(r_buf, size, MPI_CHAR, 0, i, MPI_COMM_WORLD, &stat);
            }
            MPI_Send(NULL, 0, MPI_CHAR, 0, i + 1000, MPI_COMM_WORLD);
        }
        if ( myid == 0 ) {
           double bw;
           bw = ((double)size*(double)loop)/((t_end-t_start)*(1.0e6));
           if(loopcnt != 1)
               fprintf(stdout,"Loop %d: ", l);
           fprintf(stdout,"%d\t%f (MillionBytes/sec)\t%f(MegaBytes/sec)\n",
                   size, bw, bw/1.048576);
        }
    }
    MPI_Finalize();
    return 0;
}
