/*
  Copyright (c) 2004,2005. PathScale, Inc. All rights reserved.
    
  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions
  are met:
    
  + All copies of source code must retain the above copyright notice, 
    this list of conditions and the following disclaimer. 

  + Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

  + If user makes modifications to this software, and provides those 
    modifications to PathScale, user assigns to PathScale, Inc. ownership
    of the provided modifications and all intellectual property rights 
    embodied in those modifications. 
    
  + The name of PathScale, Inc. may not be used to endorse or promote 
    products derived from this software without specific prior written 
    permission. 
      
  THIS SOFTWARE IS PROVIDED BY PATHSCALE, INC. AND ITS LICENSORS "AS IS" 
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL PATHSCALE, INC. OR ITS 
  LICENSORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
*/
/*
 * Derived from mpi_latency.c
 * Copyright (C) 2002-2003 the Network-Based Computing Laboratory
 * (NBCL), The Ohio State University.  
 *  http://nowlab.cis.ohio-state.edu/projects/mpi-iba/
 */

/* This is very similar to mpi_latency, but can run on many nodes, forwarding
 * the message around a ring of all nodes in the job.
*/

#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define MYBUFSIZE 4*1024*1025

char s_buf[MYBUFSIZE];
char r_buf[MYBUFSIZE];
int skip = 10;

int main (int argc, char *argv[])
{
    int myid, numprocs, i;
    int size;
    int loop;
    MPI_Status stat;
    double t_start = 0.0, t_end = 0.0;

    MPI_Init (&argc, &argv);
    MPI_Comm_size (MPI_COMM_WORLD, &numprocs);
    MPI_Comm_rank (MPI_COMM_WORLD, &myid);
    if (argc < 3) {
        fprintf (stderr, "Usage: %s loop msg_size [on 2 or more nodes]\n", argv[0]);
        MPI_Finalize ();
        return 0;
    }
    size = atoi (argv[2]);
    loop = atoi (argv[1]);
    if (size > MYBUFSIZE) {
        fprintf (stderr, "Maximum message size is %d\n", MYBUFSIZE);
        MPI_Finalize ();
        return 0;
    }
    /* touch the data */
    for (i = 0; i < size; i++) {
        if (myid == 0) {
            s_buf[i] = 'a';
            r_buf[i] = 'b';
        } else {
            s_buf[i] = 'B';
            r_buf[i] = 'A';
        }
    }

    MPI_Barrier (MPI_COMM_WORLD);
    if (myid == 0) {
        for (i = 0; i < skip; i++) {
            MPI_Send (s_buf, size, MPI_CHAR, 1, i, MPI_COMM_WORLD);
            MPI_Recv (r_buf, size, MPI_CHAR, numprocs - 1, i,
                      MPI_COMM_WORLD, &stat);
        }
        t_start = MPI_Wtime ();
        for (i = 0; i < loop; i++) {
            MPI_Send (s_buf, size, MPI_CHAR, 1, i, MPI_COMM_WORLD);
            MPI_Recv (r_buf, size, MPI_CHAR, numprocs - 1, i,
                      MPI_COMM_WORLD, &stat);

        }
        t_end = MPI_Wtime ();
    } else {
        for (i = 0; i < skip; i++) {
            MPI_Recv (r_buf, size, MPI_CHAR, myid - 1, i, MPI_COMM_WORLD,
                      &stat);
            MPI_Send (s_buf, size, MPI_CHAR, (myid + 1) % numprocs, i,
                      MPI_COMM_WORLD);
        }
        for (i = 0; i < loop; i++) {
            MPI_Recv (r_buf, size, MPI_CHAR, myid - 1, i, MPI_COMM_WORLD,
                      &stat);
            MPI_Send (s_buf, size, MPI_CHAR, (myid + 1) % numprocs, i,
                      MPI_COMM_WORLD);
        }
    }
    if (myid == 0) {
        double latency;

        latency = (t_end - t_start) * 1.0e6 / (numprocs * loop);
        fprintf (stdout, "%d\t%f\t\n", size, latency);
    }
    MPI_Finalize ();
    return 0;
}
