/* Copyright (c) 2006 PathScale, Inc.  All rights reserved. */


/*
 * OSU MPI Bandwidth test v2.2
 */
/*
 * Copyright (C) 2002-2006 the Network-Based Computing Laboratory
 * (NBCL), The Ohio State University. 
 *
 * Contact: Dr. D. K. Panda (panda@cse.ohio-state.edu)
 */

/*
This program is available under BSD licensing.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

(1) Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

(2) Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

(3) Neither the name of The Ohio State University nor the names of
their contributors may be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "mpi.h"
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define MAX_REQ_NUM 1000
#define MAX_PROCS 128
#define HOST_NAME_MAX 32

#define MAX_ALIGNMENT 16384
#define MAX_MSG_SIZE (1<<22)
/*N_HALF_ERROR_BOUND is maximum error in % of peak bandwidth for the N/2 
  search. */
#define N_HALF_ERROR_BOUND 0.2
#define MY_BUF_SIZE (MAX_MSG_SIZE + MAX_ALIGNMENT)
#define MANUAL 0
#define AUTOMATIC 1
struct pair
{
  int sender;
  int receiver;
}; 

int loop = 100;
int window_size = 64;
int skip = 10;

int loop_large = 20;
int window_size_large = 64;
int skip_large = 2;

int large_message_size = 8192;

char s_buf1[MY_BUF_SIZE];
char r_buf1[MY_BUF_SIZE];
struct pair pair_list[MAX_PROCS];

void usage();
double find_bw(int size, int num_pairs, char *s_buf, char *r_buf);
void find_n_half(double maxbw, int num_pairs, char *s_buf,
                 char *r_buf, int myid);
int set_pair_list(int num_pairs);

MPI_Request request[MAX_REQ_NUM];
MPI_Status  reqstat[MAX_REQ_NUM];

int main(int argc, char *argv[])
{

  int myid, num_procs, num_pairs=0, i;
  int size, align_size, org_method = AUTOMATIC;
  char *s_buf, *r_buf;
  double maxbw=0.0, bw=0.0;
  
  MPI_Init(&argc, &argv);
  for (i = 0; i < MAX_PROCS; i++)
  {
    /*initialize to -1 because it is outside the legal range for ranks*/
    pair_list[i].sender = -1;
    pair_list[i].receiver = -1;
  }
  MPI_Comm_size (MPI_COMM_WORLD, &num_procs);
  MPI_Comm_rank (MPI_COMM_WORLD, &myid);

  while ( (i = getopt(argc, argv, "-hm:")) != -1)
  {
    switch(i)
    {
      case 'h':
        if (myid == 0)
        {
          usage();
          MPI_Abort(MPI_COMM_WORLD, 0);
        }
        break;
      case 'm':
        org_method = MANUAL;
        num_pairs = (int)strtoul(optarg, NULL, 0);
        if ((num_pairs < 1))
        {
          if (myid == 0) {
            fprintf(stderr, "procs/node must be >= 1\n");
            usage();
          }
          MPI_Abort(MPI_COMM_WORLD, 1);
        } 
        break;
      default:
        if ((!optopt) && (!num_pairs) && (strtoul(optarg,NULL,0)))
        {
          num_pairs = (int) strtoul(optarg, NULL, 0);
        }         
        else
        { 
          if (myid == 0)
          {
            usage();
            MPI_Abort(MPI_COMM_WORLD, 1);
          }
        }
     }
  }    

  if(org_method == AUTOMATIC)
  {
    num_pairs = set_pair_list(num_pairs);
  }
  else
  {
    /*do the simple way */
    for (i = 0; i < num_pairs; i++)
    {
      pair_list[i].sender = i;
      pair_list[i + num_pairs].sender = i;
      pair_list[i].receiver = i + num_pairs; 
      pair_list[i + num_pairs].receiver = i + num_pairs;
    } 
    for (i = num_pairs*2; i < num_procs; i++)
    {
      pair_list[i].sender = -1;
      pair_list[i].receiver = -1;
    }
  }
  align_size = getpagesize();
  s_buf =
      (char *) (((unsigned long) s_buf1 + (align_size - 1)) /
                align_size * align_size);
  r_buf =
      (char *) (((unsigned long) r_buf1 + (align_size - 1)) /
                align_size * align_size);
  if (myid == 0) 
  {
    fprintf(stdout, "# PathScale Modified OSU MPI Bandwidth Test\n");
    fprintf(stdout, "(OSU Version 2.2, PathScale $Revision$)\n");
    fprintf(stdout, "# Running on %d procs per node\n",num_pairs);
    fprintf(stdout, "# Size\t\tAggregate Bandwidth (MB/s)\tMessages/s \n");
  }
  MPI_Barrier (MPI_COMM_WORLD);
    /* Loop over bandwidth test */
  for (size = 1; size <= MAX_MSG_SIZE; size *= 2)
  {
    /*Pulled out message test code into function find_bw*/
    bw=find_bw(size, num_pairs, s_buf, r_buf);
    if (myid == 0)
    {
      double rate;
      rate = 1000*1000*bw/size;
      fprintf(stdout, "%d\t\t%f\t\t\t%f\n", size, bw, rate);
      fflush(stdout);
      if(bw>maxbw) /*keep track of maximum bandwidth for N/2*/
      {
        maxbw=bw;
      }
    }
  }
  MPI_Bcast(&maxbw,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
  find_n_half(maxbw, num_pairs, s_buf, r_buf, myid);
  MPI_Finalize();
  return 0;
}

void usage()
{
  extern char *__progname;
  char *cmdname = strrchr(__progname, '/');
  cmdname = strrchr(__progname, '/');
  if(cmdname)
    cmdname++;
  else
    cmdname = __progname;
  printf("Usage: mpirun -np <procs> %s <num_pairs>\n", cmdname);
  printf("Unless the -m switch is used, %s requires you to run on only\n",
         cmdname);
  printf("two separate nodes.  It will pass messages between the two \n");
  printf("nodes with <num_pairs> processes on the first sending to\n");
  printf("the same number on the second.\n\n");
  printf("To run anything other than 1 node to 1 node, use the -m \n");
  printf("switch:\n");
  printf("mpirun -np <procs> %s -m <num_pairs>\n", cmdname);
  printf("This tells %s that you are (m)anually arranging the processes\n",
         cmdname);
  printf("in the hostfile or using -ppn, and will let you do more \n");
  printf("elaborate tests like 1 8-core node sending to 2 4-core nodes.\n\n");
  printf("In this case the first <num_pairs> processes by rank will send\n");
  printf("to the next <num_pairs> processes by rank.\n");
  return;
}
double find_bw(int size, int num_pairs, char *s_buf, char *r_buf)
{
/*This function is the bandwidth test that was previously part of main in 
  osu_bw, with the additional modification of being able to stream multiple
  stream multiple processors per node.  As of the InfiniPath 1.3 release
  the code can also dynamically determine which processes are on which
  node and set things up appropriately.*/
  double t_start = 0.0, t_end = 0.0, t = 0.0, maxtime=0.0, bw=0.0;
  int i, j,myid, target;
  for (i = 0; i < size; i++)
  {
    s_buf[i] = 'a';
    r_buf[i] = 'b';
  }
  
  if (size > large_message_size)
  {
    loop = loop_large;
    skip = skip_large;
    window_size = window_size_large;
  }
  MPI_Comm_rank (MPI_COMM_WORLD, &myid);
  MPI_Barrier (MPI_COMM_WORLD);
  if (pair_list[myid].sender == myid)
  {
    target = pair_list[myid].receiver;
    for (i = 0; i < loop + skip; i++)
    {
      if (i == skip)
      {
        MPI_Barrier (MPI_COMM_WORLD);
        t_start = MPI_Wtime();
        MPI_Barrier (MPI_COMM_WORLD);
      }
      for (j = 0; j < window_size; j++)
      {
        MPI_Isend (s_buf, size, MPI_CHAR, target, 100,
                  MPI_COMM_WORLD, request + j);
      }
      MPI_Waitall (window_size, request, reqstat);
      MPI_Recv (r_buf, 4, MPI_CHAR, target, 101, MPI_COMM_WORLD,
               &reqstat[0]);
    }
    t_end = MPI_Wtime();
    t = t_end - t_start;
  }
  else if (pair_list[myid].receiver == myid)
  {
    target = pair_list[myid].sender;
    for (i = 0; i < loop + skip; i++)
    {
      if (i == skip)
      {
        MPI_Barrier (MPI_COMM_WORLD);
        MPI_Barrier (MPI_COMM_WORLD);
      }
      for (j = 0; j < window_size; j++)
      {
        MPI_Irecv (r_buf, size, MPI_CHAR, target, 100,
                  MPI_COMM_WORLD, request + j);
      }
      MPI_Waitall (window_size, request, reqstat);
      MPI_Send (s_buf, 4, MPI_CHAR, target, 101, MPI_COMM_WORLD);
    }
  }
  else
  {
    MPI_Barrier(MPI_COMM_WORLD);
    MPI_Barrier(MPI_COMM_WORLD);
  }
  MPI_Reduce (&t, &maxtime, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
  if (myid==0)
  {
    double tmp;
    tmp = ( (num_pairs * size * 1.0) / (1000 * 1000) );
    tmp = tmp * loop * window_size;
    bw = tmp / maxtime;
    return bw;
  }
  return 0;
}
void find_n_half(double maxbw, int num_pairs, char *s_buf, 
                 char *r_buf, int myid)
{
  double bw = 0.0, error;
  int size = 1, lastsize = 0, top=0, bottom=0, tmp;
  error = (maxbw / 100) * N_HALF_ERROR_BOUND;
  if (myid == 0)
  {
    fprintf(stdout, "Searching for N/2 bandwidth.  Maximum ");
    fprintf(stdout, "Bandwidth of %f MB/s", maxbw);
    fprintf(stdout, "...\n");
    fflush(stdout);
  }
  while ( (size != lastsize) && 
        (bw < (maxbw / 2 - error) || bw > (maxbw / 2 + error)))
  {
    bw = find_bw (size, num_pairs, s_buf, r_buf); 
    if (myid == 0)
    {
      if (bw < (maxbw / 2 - error))
      {
        tmp = size;
        if (lastsize > size || top) /*We have a top! */
        {
          if (lastsize > size) 
          {
            top = lastsize;
          }
          size = size + (top - size) / 2;
        }
        else /*Still searching for a top */
        {
          size = size << 1 ;
        }
        lastsize = tmp;
      }
      if( bw > (maxbw / 2 + error))
      {
        tmp = size;
        if (lastsize < size)
        {
          bottom = lastsize;
        }
        size = size - (size - bottom) / 2;
        lastsize = tmp;
      }
    }
    MPI_Bcast (&size, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast (&lastsize, 1, MPI_INT, 0, MPI_COMM_WORLD);
    MPI_Bcast (&bw, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  }
  if (myid == 0)
  {
    fprintf (stdout, "Found N/2 bandwidth of %f MB/s at size",bw);
    fprintf (stdout, " %i bytes\n", size); 
  }
  MPI_Barrier (MPI_COMM_WORLD);
  return;
}

int set_pair_list (int num_pairs)
/* pair_list is global.  Also note every rank is doing this, so most errors *
 * will be seen by all ranks.  Thus for many of these it should be safe to  *
 *  have only the first call usage and abort. */
{
  char host_name[HOST_NAME_MAX], host_name_buff[HOST_NAME_MAX*MAX_PROCS];
  char host[2][HOST_NAME_MAX];
  size_t len = HOST_NAME_MAX;
  int num_procs, myid, i, j;
  int proc_counter[2] = {0,0};
  
  MPI_Comm_size (MPI_COMM_WORLD, &num_procs);
  MPI_Comm_rank (MPI_COMM_WORLD, &myid);
  memset (host, 0, HOST_NAME_MAX*2);
  memset (host_name, 0, HOST_NAME_MAX);
  memset (host_name_buff, 0, HOST_NAME_MAX*MAX_PROCS);
  if (gethostname (host_name, len))
  {
    fprintf (stderr, "Error with rank %d getting hostname!\n", myid);
    MPI_Abort (MPI_COMM_WORLD, 1);
  }
  else
  { 
    MPI_Allgather(host_name, HOST_NAME_MAX, MPI_CHAR, host_name_buff, 
                  HOST_NAME_MAX, MPI_CHAR, MPI_COMM_WORLD);
    for (i = 0; i < num_procs; i++)
      {
      /*If the first host is not null and they are not identical */
      if (*host[0] && (strncmp (host[0], &host_name_buff[i*HOST_NAME_MAX],
                                         HOST_NAME_MAX) != 0))
      {
        /*And the second is not null and they are not identical */
        if (*host[1] && (strncmp (host[1], &host_name_buff[i*HOST_NAME_MAX],
                                          HOST_NAME_MAX) != 0))
        {
          if (myid == 0)
          {
            usage();
            fprintf (stderr, "\nMore than 2 hosts identified! Aborting!\n");
            MPI_Abort (MPI_COMM_WORLD,1);
          }
        }
        else 
        {
          /*second host is either null or identical, so can assign */
          memcpy (host[1], &host_name_buff[i*HOST_NAME_MAX], HOST_NAME_MAX);
          /*on second host, so assign to be receiver*/
          pair_list[i].receiver = i;
          /*If num_pairs not specified, don't check this */
          if ( (++proc_counter[1] > num_pairs) && num_pairs)
          {
            if (myid == 0)
            {
              fprintf (stderr, "\nMore than specified %d ", num_pairs);
              fprintf (stderr, "processes found on node 1!  Aborting!\n");
              MPI_Abort (MPI_COMM_WORLD, 1); 
            }
          } 
        }
      }
      else /*first host is either null or identical, so can assign */
      {
        memcpy (host[0], &host_name_buff[i*HOST_NAME_MAX], HOST_NAME_MAX);
        /*on first host, so assign to be sender*/
        pair_list[i].sender = i;
        /*If procs per node not specified, don't check this */
        if ( (++proc_counter[0] > num_pairs) && num_pairs)
        {
          if (myid == 0)
          {
            fprintf (stderr, "\nMore than specified %d ", num_pairs);
            fprintf (stderr, "processes found on node 0!  Aborting!\n");
            MPI_Abort(MPI_COMM_WORLD, 1);
          }
        }
      }
    } /*end for (i = 0; i < num_procs; i++) */
    if ( strncmp (host[1], "", 1) == 0 )  //if we have only 1 node
    {
      if (myid == 0)
      {
        fprintf (stderr, "Only one node detected! Aborting\n");
        MPI_Abort (MPI_COMM_WORLD, 1);
      }
    }
    if (proc_counter[0] != proc_counter[1])
    {
      if (myid == 0)
      {
        fprintf (stderr, "Number of processes different on node 0 than\n");
        fprintf (stderr, "on node one!  Aborting!\n");
        MPI_Abort (MPI_COMM_WORLD, 1);
      }
    }
    if (num_pairs == 0) 
    {
      num_pairs = proc_counter[0];
    }
    if (num_pairs > proc_counter[0] || 
        num_pairs > proc_counter[1])
    {
      if (myid == 0)
      {
        fprintf (stderr, "There aren't enough processes on the nodes to \n");
        fprintf (stderr, "run with your specified num_pairs: %i.  Aborting!\n", 
                 num_pairs);
        MPI_Abort (MPI_COMM_WORLD, 1);
      }
    }
        
    /*If we got here we know we have 2 nodes, and every element of pair_list
     *up to num_procs - 1 has either a sender or a receiver.  We want to 
     *walk through and match them all up. */
    for (i = 0; i < num_procs; i++)
    {
      if (pair_list[i].receiver < 0)
      /*If we don't know a receiver*/
      { 
        for (j = 0; j < num_procs; j++)
        /*Loop over the whole list so that we don't have to separately do
        * receiver case.  */
        {
          if (pair_list[j].sender < 0)
          /*Find someone who doesn't know a receiver*/ 
          {
            pair_list[j].sender = i;
            pair_list[i].receiver = j;
            break;
          }
        }
      }
    } /*Should now be all nice and paired up.*/
  }  /* end else of if gethostbyname */
  /* We return num_pairs, in case we figured it out for main */
  return num_pairs;
} 
