#!/bin/bash
# Copyright (C) 2006 QLogic Corporation, All rights reserved.
# Copyright 2005, 2006, PathScale, Inc. All rights reserved.
#
# Permission is granted to PathScale customers to use this software in 
# source and binary forms, with or without modification, provided that 
# the following conditions are met: 
# 
# + All copies of source code must retain the above copyright notice, 
#   this list of conditions and the following disclaimer. 
# 
# + Customer has a valid license agreement from PathScale, Inc. for the 
#   software product with which this software is originally distributed. 
# 
# + Redistribution of original or modified versions to third parties is 
#   prohibited without specific prior written permission of PathScale, 
#   Inc. 
# 
# + If customer makes modifications to this software, and provides those 
#   modifications to PathScale, customer assigns to PathScale, Inc. 
#   ownership of the provided modifications and all intellectual property 
#   rights embodied in those modifications. 
# 
# + The name of PathScale, Inc. may not be used to endorse or promote 
#   products derived from this software without specific prior written 
#   permission. 
# 
# THIS SOFTWARE IS PROVIDED BY PATHSCALE, INC. AND ITS LICENSORS "AS IS" 
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL PATHSCALE, INC. OR ITS 
# LICENSORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

# Make sure we have a sane default PATH
export PATH=/sbin:/bin:/usr/sbin:/usr/bin:$PATH

KPREFIX="${KPREFIX:-/lib/modules}"
KVER="${KVER:-$(uname -r)}"
KDIR="${KDIR:-${KPREFIX}/${KVER}}"
KBDIR="${KBDIR:-${KDIR}/build}"

if [ "x$RPM_INSTALL_PREFIX" = x ]; then
    RPM_INSTALL_PREFIX=/usr
fi

cd $RPM_INSTALL_PREFIX/src/infinipath/drivers
source ./build-info.sh
source ./build-guards.sh
    
SRCROOT=`pwd`/drivers-$distro
SRC=$SRCROOT/drivers/infiniband

export CONFIG_INFINIBAND=m
export CONFIG_INFINIBAND_ADDR_TRANS=y
export CONFIG_INFINIBAND_EHCA=n
export CONFIG_INFINIBAND_IPATH=$CONFIG_INFINIBAND
export CONFIG_INFINIBAND_IPOIB=$CONFIG_INFINIBAND
export CONFIG_INFINIBAND_ISER=n
export CONFIG_INFINIBAND_MADEYE=$CONFIG_INFINIBAND
export CONFIG_INFINIBAND_MTHCA=$CONFIG_INFINIBAND
export CONFIG_INFINIBAND_SDP=$CONFIG_INFINIBAND
export CONFIG_INFINIBAND_SRP=$CONFIG_INFINIBAND
export CONFIG_INFINIBAND_USER_ACCESS=$CONFIG_INFINIBAND
export CONFIG_INFINIBAND_USER_MAD=$CONFIG_INFINIBAND
export CONFIG_IPATH_ETHER=$CONFIG_INFINIBAND_IPATH
export CONFIG_KDAPL=n
export CONFIG_KDAPL_INFINIBAND=n

remove_modules() {
    rm -f \
        ${KDIR}/kernel/drivers/net/ib_ipath.ko \
        ${KDIR}/kernel/drivers/net/infinipath.ko \
        ${KDIR}/kernel/drivers/net/ipath_core.ko \
        ${KDIR}/kernel/drivers/net/ipath_ether.ko \
        ${KDIR}/updates/findex.ko \
        ${KDIR}/updates/ib_addr.ko \
        ${KDIR}/updates/ib_at.ko \
        ${KDIR}/updates/ib_cm.ko \
        ${KDIR}/updates/ib_core.ko \
        ${KDIR}/updates/ib_ipath.ko \
        ${KDIR}/updates/ib_ipoib.ko \
        ${KDIR}/updates/ib_local_sa.ko \
        ${KDIR}/updates/ib_mad.ko \
        ${KDIR}/updates/ib_mthca.ko \
        ${KDIR}/updates/ib_ping.ko \
        ${KDIR}/updates/ib_sa.ko \
        ${KDIR}/updates/ib_sdp.ko \
        ${KDIR}/updates/ib_srp.ko \
        ${KDIR}/updates/ib_uat.ko \
        ${KDIR}/updates/ib_ucm.ko \
        ${KDIR}/updates/ib_umad.ko \
        ${KDIR}/updates/ib_uverbs.ko \
        ${KDIR}/updates/ib_madeye.ko \
        ${KDIR}/updates/ipath_core.ko \
        ${KDIR}/updates/ipath_ether.ko \
        ${KDIR}/updates/rdma_cm.ko \
        ${KDIR}/updates/rdma_ucm.ko
}

if [ "$1" == clean ]; then
    if [ -d $SRC ]; then
	make -s -C ${KBDIR} M=${SRC} CC="gcc -m64 -I$SRC/../../include" clean
    fi
    remove_modules
    if [ -d $SRCROOT ]; then
        rm -rf $SRCROOT
    fi
    exit 0
fi

UDEV_STATIC=/etc/udev/static_devices.txt

clean_udev_static()
{
    local tmp
    tmp=`mktemp`
    grep -v '^ipath' ${UDEV_STATIC} > ${tmp}
    cmp -s ${UDEV_STATIC} ${tmp} || cp ${tmp} ${UDEV_STATIC}
    rm ${tmp}
}

clean_sles9_rules()
{
    local tmp
    tmp=`mktemp`
    egrep -v '^KERNEL="((ucm|uat|ucma|rdma_cm|issm|umad|uverbs)\*|ipath.*)"' $1 > ${tmp}
    cmp -s $1 ${tmp} || cp ${tmp} $1
    rm ${tmp}
}

clean_sles9_permissions()
{
    local tmp
    tmp=`mktemp`
    egrep -v '^(ipath|infiniband)' $1 > ${tmp}
    cmp -s $1 ${tmp} || cp ${tmp} $1
    rm ${tmp}
}

cat_permissions()
{
    local perms
    perms=$1

    echo 'ipath:root:root:0666' >> ${perms}
    echo 'ipath[0-9]*:root:root:0666' >> ${perms}
    echo 'ipath_diag[0-9]*:root:root:0600' >> ${perms}
    echo 'ipath_diagpkt:root:root:0600' >> ${perms}
    echo 'infiniband/umad*:root:root:0600' >> ${perms}
    echo 'infiniband/issm*:root:root:0600' >> ${perms}
    echo 'infiniband/uverbs*:root:root:0666' >> ${perms}
    echo 'infiniband/ucm*:root:root:0666' >> ${perms}
    echo 'infiniband/rdma_cm*:root:root:0666' >> ${perms}
    chmod 644 ${perms}
}

cat_rules()
{
    local rules
    rules=$1
    local keq
    # more recent versions of udev (e.g. FC4) want slightly
    # different syntax.
    if grep -qsl '^KERNEL==' ${ruledir}/50-udev*.rules; then
	keq="KERNEL=="
    else
	keq="KERNEL="
    fi
    echo ${keq}'"ipath", NAME="%k", MODE="0666"' >> ${rules}
    echo ${keq}'"ipath[0-9]*", NAME="%k", MODE="0666"' >> ${rules}
    echo ${keq}'"ipath_*", NAME="%k", MODE="0600"' >> ${rules}
    echo ${keq}'"umad*", NAME="infiniband/%k"' >> ${rules}
    echo ${keq}'"issm*", NAME="infiniband/%k"' >> ${rules}
    echo ${keq}'"ucm*", NAME="infiniband/%k", MODE="0666"' >> ${rules}
    echo ${keq}'"uverbs*", NAME="infiniband/%k", MODE="0666"' >> ${rules}
    echo ${keq}'"uat*", NAME="infiniband/%k", MODE="0666"' >> ${rules}
    echo ${keq}'"ucma*", NAME="infiniband/%k", MODE="0666"' >> ${rules}
    echo ${keq}'"rdma_cm*", NAME="infiniband/%k", MODE="0666"' >> ${rules}
    chmod 644 ${rules}
}

mk_udev_dynamic()
{
    local ruledir
    if [ -d /etc/udev/rules.d ]; then
       ruledir=/etc/udev/rules.d
       rm -f ${ruledir}/[46]0-ipath.rules
    elif [ "$distro" == "sles9sp3" ]; then
       ruledir=/etc/udev
       clean_sles9_rules ${ruledir}/udev.rules
    else
       echo '*** ERROR: could not determine how to configure udev!' 1>&2
       echo '*** Could not locate rules files.' 1>&2
    fi
    local permdir
    local perms
    if [ -d /etc/udev/permissions.d ]; then
       permdir=/etc/udev/permissions.d
       perms=${permdir}/40-ipath.permissions
       rm -f ${permdir}/[46]0-ipath.permissions
    elif [ "$distro" == "sles9sp3" ]; then
       permdir=/etc/udev
       perms=${permdir}/udev.permissions
       clean_sles9_permissions ${perms}
    fi

    local ctrl
    ctrl=/sbin/udevcontrol

    if [ -f ${permdir}/50-udev.permissions ]; then
	rm -f ${perms}
	cat_permissions ${perms}
    elif [ -f /etc/udev/udev.permissions ]; then
	cat_permissions ${perms}
    fi

    if [ -f ${ruledir}/50-udev.rules -o -f ${ruledir}/50-udev-default.rules ]; then
	local rules
	# it seems that redhat/fedora really wants the ipath rules
	# file to come after the udev default rules in lexicographic
	# order, while suse wants them before.
	if [ -e /etc/redhat-release -o -e /etc/fedora-release ]; then
	    rules=${ruledir}/60-ipath.rules
	else
	    rules=${ruledir}/40-ipath.rules
	fi
	
	rm -f ${rules}
	cat_rules ${rules}
	[ -x ${ctrl} ] && ${ctrl} reload_rules > /dev/null 2>&1
    elif [ -f ${ruledir}/udev.rules ]; then
	cat_rules ${ruledir}/udev.rules
    else
	echo '*** WARNING: could not determine how to configure udev!' 1>&2
	echo '*** /dev/ipath* files may have incorrect permissions!' 1>&2
    fi
}

clean_ipath_dev()
{
    rm -f $1/ipath $1/ipath[0-9]* $1/ipath_ctrl $1/ipath_diag \
	$1/ipath_diagpkt $1/ipath_diag[0-9]*
    rm -rf $1/infiniband
}

rm -rf $SRCROOT
tar mjxf sources/drivers-$distro.tar.bz2
ret=$?
if [ $ret != 0 ]; then
    echo "*** Unpack of source tarball failed: exit status $ret" 1>&2
    exit $ret
fi

echo '  'Building and installing InfiniPath modules for ${distro} ${KVER} kernel
LD_PRELOAD=$L make -s -C ${KBDIR} M=${SRC} CC="gcc -m64 -I$SRC/../../include"
ret=$?
if [ $ret != 0 ]; then
    echo "*** Driver build failed: exit status $ret" 1>&2
    exit $ret
fi

mkdir -p ${KDIR}/updates

d=${KDIR}/kernel/drivers/infiniband/hw/ipath
if [ -f $d/ipath_core.ko ]; then
    mv $d/ipath_core.ko $d/ipath_core.ko.bak
fi

remove_modules

find $SRC -name \*.ko -exec install -m 744 {} ${KDIR}/updates \;
/sbin/depmod -ae ${KVER}
ret=$?
if [ $ret != 0 ]; then
    echo "*** depmod step failed: exit status $ret" 1>&2
    exit $ret
fi

if [ -e /etc/redhat-release ]; then
	MPDC=/etc/modprobe.conf
else
	MPDC=/etc/modprobe.conf.local
fi

# remove the line added by infinipath 1.3 up through 2.0Beta1
# if present exactly as originally added.  It's no longer needed
# and can confuse people when modprobe options are ignored
if grep -q '^install ib_ipath' $MPDC; then
	ed -s -p '' $MPDC << \EOF
g,^install ib_ipath /sbin/modprobe ib_mad; /sbin/modprobe --ignore-install ib_ipath$,d
w
q
EOF
fi

if [ -d /etc/udev ]; then
	mk_udev_dynamic
	# older udev just make devices in /etc/udev/devices
	# newer udev (as in SuSE 10), have to add to ${UDEV_STATIC})
	if [ -e ${UDEV_STATIC} ]; then
		clean_udev_static
	else
		clean_ipath_dev /etc/udev/devices
	fi
fi

exit 0
